import pathlib

import click
import requests

import epcsunspecdemo.clishared


smdx_file_template = 'smdx_{model_index:05}.xml'

url_template = (
    'https://raw.githubusercontent.com'
    '/epcpower/epcpower.github.io/master/release_artifacts'
    '/{version}'
    '/' + smdx_file_template
)


@click.command(
    name='get-models',
    help='Download EPC SMDX model definitions',
)
@click.option(
    '--version',
    default='1.2.5',
    help='Converter software version',
)
@click.option(
    '--model-index',
    'model_indexes',
    multiple=True,
    type=int,
    default=[65534],
    help='Model number to download',
)
@epcsunspecdemo.clishared.model_path_option
def get_models(version, model_indexes, model_path):
    version = version.split('.')

    model_path = pathlib.Path(model_path)
    model_path.mkdir(parents=True, exist_ok=True)

    for model_index in model_indexes:
        url = url_template.format(
            version='_'.join(version),
            model_index=model_index,
        )

        smdx_file_name = smdx_file_template.format(model_index=model_index)

        response = requests.get(url)
        file = model_path/smdx_file_name
        file.write_bytes(response.content)
        click.echo('Saved {}'.format(smdx_file_name))



import decimal
import sys

import click

import epcsunspecdemo.clishared


@click.group(
    name='write',
    help='Write data points',
)
def group():
    pass


converters = {
    'int16': decimal.Decimal,
    'uint16': decimal.Decimal,
    'int32': decimal.Decimal,
    'uint32': decimal.Decimal,
    'sunssf': None,
    'enum16': decimal.Decimal,
    'bitfield16': int,
    'bitfield32': int,
    'string': str,
    'acc16': None,
    'acc32': None,
    'pad': None,
}


def common(
        device_factory,
        model_name,
        point_names_and_values,
        repeating_point_names_and_values,
):
    device = device_factory()
    model = device[model_name]

    if model is None:
        click.echo('No model found by the name {!r}'.format(model_name))

        click.echo(
            'Perhaps you meant one of: {}'.format(', '.join(device.models)),
        )

        sys.exit(1)

    model.read()

    all_points = [
        (0, name, value)
        for name, value in point_names_and_values
    ]
    all_points.extend(repeating_point_names_and_values)

    for block_index, point_name, value in all_points:
        client_point = model.model.blocks[block_index].points[point_name]
        sunspec_point_type = client_point.point_type.type

        converter = converters.get(sunspec_point_type)

        if converter is None:
            raise Exception()
        else:
            client_point.value = converter(value)
            client_point.write()


commands = epcsunspecdemo.clishared.Commands.build(
    options=(
        epcsunspecdemo.clishared.repeating_point_names_and_values_option,
        epcsunspecdemo.clishared.model_name_option,
        epcsunspecdemo.clishared.point_names_and_values_option,
    ),
    common=common,
)
commands.add_to(group)



import sys

import attr
import click
import serial.serialutil
import sunspec.core.client

import epcsunspecdemo.utils


model_path_option = click.option(
    '--model-path',
    type=click.Path(file_okay=False),
    default='models',
    help='Path to the directory containing custom smdx_*.xml files',
)


timeout_option = click.option(
    '--timeout',
    default=1,
    type=float,
    help='Modbus communication timeout in seconds',
)


if sys.platform.startswith('win'):
    port_help = 'Name of the COM port'
else:
    port_help = 'Path to the serial device'


serial_port_option = click.option(
    '--port',
    required=True,
    help=port_help,
)


serial_baudrate_option = click.option(
    '--baudrate',
    type=click.Choice(
        str(rate)
        for rate in serial.serialutil.SerialBase.BAUDRATES
    ),
    default='9600',
    help='Serial baudrate',
)


tcp_address_option = click.option(
    '--address',
    required=True,
    help='The IP address or host name of the converter',
)


tcp_port_option = click.option(
    '--port',
    type=int,
    default=502,
    help='The TCP port on the converter',
)


slave_id_option = click.option(
    '--slave-id',
    type=int,
    default=1,
    help='Node ID of the converter',
)


model_name_option = click.option(
    '--model-name',
    required=True,
    help='The name of the model',
)


point_names_option = click.option(
    '--point-name',
    'point_names',
    multiple=True,
    help='The name of the point',
)


point_names_and_values_option = click.option(
    '--point',
    'point_names_and_values',
    nargs=2,
    type=(str, str),
    multiple=True,
    help='The name of the point and the value',
)


repeating_point_names_and_values_option = click.option(
    '--repeating-point',
    'repeating_point_names_and_values',
    nargs=3,
    type=(int, str, str),
    multiple=True,
    help='The name of the point and the value',
)


@attr.s
class RtuDeviceFactory:
    port = attr.ib()
    baudrate = attr.ib()
    timeout = attr.ib()
    slave_id = attr.ib()
    model_path = attr.ib()

    def __call__(self):
        with epcsunspecdemo.utils.fresh_smdx_path(self.model_path):
            return sunspec.core.client.SunSpecClientDevice(
                slave_id=self.slave_id,
                device_type=sunspec.core.client.RTU,
                name=self.port,
                baudrate=self.baudrate,
                timeout=self.timeout,
            )


@attr.s
class TcpDeviceFactory:
    address = attr.ib()
    port = attr.ib()
    timeout = attr.ib()
    slave_id = attr.ib()
    model_path = attr.ib()

    def __call__(self):
        with epcsunspecdemo.utils.fresh_smdx_path(self.model_path):
            return sunspec.core.client.SunSpecClientDevice(
                slave_id=self.slave_id,
                device_type=sunspec.core.client.TCP,
                ipaddr=self.address,
                ipport=self.port,
                timeout=self.timeout,
            )


@attr.s
class Commands:
    rtu = attr.ib()
    tcp = attr.ib()

    @classmethod
    def build(cls, options, common):
        common_options = (
            epcsunspecdemo.clishared.timeout_option,
            epcsunspecdemo.clishared.slave_id_option,
            epcsunspecdemo.clishared.model_path_option,
        )

        @click.command()
        @epcsunspecdemo.utils.apply_decorators(options)
        @serial_port_option
        @serial_baudrate_option
        @epcsunspecdemo.utils.apply_decorators(common_options)
        def rtu(port, baudrate, timeout, slave_id, model_path, **kwargs):
            device_factory = RtuDeviceFactory(
                port=port,
                baudrate=baudrate,
                timeout=timeout,
                slave_id=slave_id,
                model_path=model_path,
            )

            common(
                device_factory=device_factory,
                **kwargs
            )

        @click.command()
        @epcsunspecdemo.utils.apply_decorators(options)
        @tcp_address_option
        @tcp_port_option
        @epcsunspecdemo.utils.apply_decorators(common_options)
        def tcp(address, port, timeout, slave_id, model_path, **kwargs):
            device_factory = TcpDeviceFactory(
                address=address,
                port=port,
                timeout=timeout,
                slave_id=slave_id,
                model_path=model_path,
            )

            common(
                device_factory=device_factory,
                **kwargs
            )

        return cls(rtu=rtu, tcp=tcp)

    def add_to(self, group):
        for command in (self.rtu, self.tcp):
            group.add_command(command)



import itertools
import sys
import time

import click
import sunspec.core.suns
import toolz
import tqdm

import epcsunspecdemo.clishared


@click.group(
    name='datalogger',
    help='Access the high speed data logger',
)
def group():
    pass


def terminate_on_model_failure(device, model, name):
    if model is None:
        click.echo('No model found by the name {!r}'.format(name))

        click.echo(
            'Perhaps you meant one of: {}'.format(', '.join(device.models)),
        )

        sys.exit(1)


def read(registers_per_read, log_model, total_octets):
    total_16_bit_bytes, remainder = divmod(total_octets, 2)

    assert log_model.model.points['R000'].point_type.type == 'uint16', 'there is no type to unimplemented value map so we just assume uint16 for now'
    unimplemented_value = sunspec.core.suns.SUNS_UNIMPL_UINT16

    offset = 0
    while offset < total_16_bit_bytes:
        log_model.Offset = offset
        log_model.model.points['Offset'].write()
        log_model.read()

        registers_to_read = min(
            registers_per_read,
            total_16_bit_bytes - offset,
        )

        for n in range(registers_to_read):
            register_value = getattr(log_model, 'R{:03}'.format(n))
            if register_value is None:
                register_value = unimplemented_value

            # the Modbus registers are 16-bits, or two local 8-bit bytes
            byte_pair = register_value.to_bytes(
                2,
                byteorder='big',
                signed=False,
            )
            for byte in byte_pair:
                yield byte

        offset += registers_to_read


def common(device_factory, registers_per_read, log_file):
    device = device_factory()

    configuration_model_name = 'data_logger'
    configuration_model = device[configuration_model_name]
    terminate_on_model_failure(
        device=device,
        model=configuration_model,
        name=configuration_model_name
    )

    log_model_name = 'data_logger_log'
    log_model = device[log_model_name]
    terminate_on_model_failure(
        device=device,
        model=log_model,
        name=log_model_name
    )

    configuration_model.read()

    while configuration_model.RdblOct == 0:
        print('Waiting for readable data...')
        time.sleep(1)
        configuration_model.read()

    total_octets = configuration_model.RdblOct

    bytes_iterable = read(
        registers_per_read=registers_per_read,
        log_model=log_model,
        total_octets=total_octets,

    )
    tqdm_iterable = tqdm.tqdm(
        iterable=bytes_iterable,
        total=total_octets,
        unit='b',
        unit_scale=True,
        unit_divisor=1024,
    )

    for block in toolz.partition_all(1024, tqdm_iterable):
        log_file.write(bytes(block))


commands = epcsunspecdemo.clishared.Commands.build(
    options=(
        click.option(
            '--registers-per-read',
            default=120,
            type=click.IntRange(min=1, max=123),
        ),
        click.option(
            '--log-file',
            required=True,
            type=click.File(mode='wb', atomic=True),
        )
    ),
    common=common,
)
commands.add_to(group)
